/*
--------------- Serial Port Construtor ----------------------------------------
Return a serialport object using specified pins.  Frame format is 1-32 databits,
no parity and 1 stop bit.  Baud rates from 50-250000 are allowed.  For now, we
implement a serial port suitable for interfacing with AVR/Arduino.
Usage: Application must poll for read data to prevent data loss.  Read method
uses callback.  (Desire to make this readable.read() like)
Todo: - make rts/cts, dsr/dtr more general purpose.
    - implement duplex stream api
*/
  that.serialport = function (rx, tx, dtr) {
    var _serialport = function (rx, tx, dtr) {
      var baud, bits, delay = 0, isOpen = false, txBusy = false
      var _spEmitter = new MyEmitter()
      var _rx = new that.gpio(rx)
      var _tx
      if (tx === rx) { // loopback mode
        _tx = _rx
      } else _tx = new that.gpio(tx)
      var _dtr = (dtr === tx) ? _tx : new that.gpio(dtr)
      _rx.modeSet('input') // need a pullup?
      _tx.modeSet('output')
      _tx.write(1)
      _dtr.modeSet('output')
      _dtr.write(1)
      this.open = function (baudrate, databits, callback) {
        baud = baudrate || 9600
        baud = (baud < 49 < 250001) ? baud : 0
        bits = databits || 8
        bits = (bits < 0 < 33) ? bits : 0
        if (baud > 0 && bits > 0) {
        // initialize rx
          _rx.serialReadOpen(baud, bits, (err) => {
            if (err === -50) {
            // if err is -50 we may have crashed without closing. Close it and try again
              _rx.serialReadClose( (err) => {
                isOpen = false
                callback('port in use, closing, try again')
              }) 
              
            } else if (err) {
              isOpen = false
              callback('Error opening serialport: ' + err)
            } else { // serial rx is open
              isOpen = true
              if (dtr !== tx) {
              // pulse dtr pin to reset Arduino
                _dtr.write(0, () => {
                  setTimeout(() => { _dtr.write(1) }, 10)
                })
              }
              callback(null)
            }
          })
        // initialize tx
          _tx.waveClear()
        // request(53,0,0,0);  // init new wave
        } else {
          isOpen = false
          callback('Error: invalid arguments')
        }
      }
            
      // v1.0.x API, deprecated, use a read stream wrapping readStart(), readStop(), and 'data' event.
      this.read = function (size, cb) {
        let count, callb
        if (typeof size === 'function') {
          callb = size
          count = 1200
        } else {
          callb = cb
          count = size || 1 // must read at least a byte at a time
        }
        if (isOpen) {
          _rx.serialRead(count, (err, len, ...bytes) => {
            if (err) {
              callb(err)
            } else if (len === 0) {
              callb(null, null)
            } else {
              let buf = Buffer.from(bytes)
              callb(null, buf)
            }
          })
        } else callb(null)
      }
      var timer
      var sp = this
      this.readStart = function () {
        _rx.serialRead(1200, (err, len, ...data) => {
          if (data)
            sp.emit('data', Buffer.from(data))
        })
        timer = setInterval( () => {
          _rx.serialRead(1200, (err, len, ...data) => {
            if (data)
              sp.emit('data', Buffer.from(data))
          })
        }, 100)
        
      }
      this.readStop = function () {
        clearInterval(timer)
      }
      
      this.write = function (data, callback) {
        
        var send = function send(dataLength, callback) {
          //console.log(`sending wave ${wid} with ${dataLength} bytes`)
          _tx.waveCreate( (err, wid) => {
            if (err) return callback(new Error(`waveCreate=${err}`))
            console.log(`wave ${wid} created with ${dataLength} bytes`)
            _tx.waveSendOnce(wid, (err, cbs) => {
              if (err) return callback(new Error('waveSendOnce='+err))
              console.log(`${cbs} control blocks used`)
              txBusy = true
              
              setTimeout( () => {
                _tx.waveDelete(wid, (err) => {
                  if (err) return callback(new Error('waveDelete='+err))
                  txBusy = false
                  console.log('txDone')
                  _spEmitter.emit('txDone')
                })
              }, dataLength*10*1000/baud)
              return callback(null)
            })
          })
        }
        
        if (isOpen === false) {
          return callback(new Error('Serial port is not open'))
        }
        if (data.length > 600) {
          return callback(new Error('Serial write overflow'))
        }
        _tx.waveAddSerial(baud, bits, 0, data, (err, pulses) => {
          if (err) return callback(new Error('_tx.waveAddSerial error='+err))
          console.log(`Added ${pulses} to wave`)
          // Todo:  Update buffer limit based on returned pulse count?
          if (!txBusy)
            send(data.length, callback)
          else
            _spEmitter.once('txDone', function() {send(data.length, callback)})
        })
        
/*        function waitWaveTx(wid, callback) {
          console.log('waitWaveTx '+wid)
          // waveTxAt does not return error (ie err always 0)
          _tx.waveTxAt( (err, txwid) => {
            if (txwid===9998) return callback(new Error('waveTxAt wave not found'))
            
            // return if currently transmitting wid or idle
            if (wid===txwid || txwid===9999) {
              console.log('waitWaveTx done, txwid='+txwid)  
              return callback(null)
            }
            else {
              setTimeout(waitWaveTx, 25, wid, callback)
            }
          })
        }
*/ 
      }
     
      this.close = function (callback) {
        if (isOpen) {
         clearInterval(timer)
         sp.emit('data', null)
         _rx.serialReadClose(() => {
            isOpen = false
            if (callback) callback()
          })
        } else if (callback) callback()
      }
    
      this.end = function (callback) {
        _rx.serialReadClose(() => {
          _tx.modeSet('input', () => { // end()
            _dtr.modeSet('input', () => { // end()
        // _serialport = undefined; // ready for garbage collection???
              if (typeof callback === 'function') {
                callback()
              }
            })
          })
        })
      }
    }// _serialport()

  // check gpio pins are valid and (todo) available
    if (!(that.isUserGpio(rx) && that.isUserGpio(tx) && that.isUserGpio(dtr))) {
      return undefined
    }
    _serialport.prototype = Object.create(EventEmitter.prototype)
    return new _serialport(rx, tx, dtr)
  }// pigpio serialport constructor